/***********************************************************************************
* Run Command: Plasmoid to run commands without terminal.
* Copyright (C) 2008 - 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "RunCommand.h"

#include <QGraphicsLinearLayout>
#include <QScriptEngine>
#include <QStringList>
#include <QFileInfo>
#include <QRegExp>
#include <QDir>

#include <Plasma/Theme>

#include <kworkspace/kdisplaymanager.h>
#include <kworkspace/kworkspace.h>
#include <KServiceTypeTrader>
#include <KToolInvocation>
#include <KGlobalSettings>
#include <KProtocolInfo>
#include <KStandardDirs>
#include <KCompletion>
#include <KAuthorized>
#include <KLineEdit>
#include <KService>
#include <KShell>
#include <KUrl>
#include <KRun>

K_EXPORT_PLASMA_APPLET(runcommand, RunCommand)

RunCommand::RunCommand(QObject *parent, const QVariantList &args) : Plasma::Applet(parent, args)
{
    resize(200, 50);
}

void RunCommand::init()
{
    KSharedConfig::Ptr sharedConfiguration = KSharedConfig::openConfig("krunnerrc");
    KConfigGroup *configuration = new KConfigGroup(sharedConfiguration, "General");

    QStringList commands = configuration->readEntry("PastQueries", QStringList());

    m_lastRun = QDateTime::currentDateTime();

    KHistoryComboBox *comboBox = new KHistoryComboBox(false);
    comboBox->setAttribute(Qt::WA_NoSystemBackground);
    comboBox->setCompletionMode(KGlobalSettings::CompletionAuto);
    comboBox->completionObject()->insertItems(commands);
    comboBox->completionObject()->setOrder(KCompletion::Sorted);
    comboBox->setHistoryItems(commands);
    comboBox->setEditable(true);
    comboBox->clearEditText();

    m_comboBox = new QGraphicsProxyWidget(this);
    m_comboBox->setWidget(comboBox);

    QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(this);
    layout->setContentsMargins(3, 0, 3, 0);
    layout->setSpacing(0);
    layout->addItem(m_comboBox);

    setLayout(layout);

    setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed));

    setMaximumSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX);

    setMinimumSize(0, 0);

    constraintsEvent(Plasma::FormFactorConstraint);

    connect(this, SIGNAL(activate()), this, SLOT(activated()));
    connect(comboBox, SIGNAL(cleared()), this, SLOT(clearHistory()));
    connect(comboBox, SIGNAL(returnPressed()), this, SLOT(getCommand()));
    connect(comboBox, SIGNAL(activated(QString)), this, SLOT(runCommand(QString)));
    connect(comboBox, SIGNAL(editTextChanged(QString)), this, SLOT(textChanged(QString)));
}

void RunCommand::constraintsEvent(Plasma::Constraints constraints)
{
    if (constraints & Plasma::FormFactorConstraint)
    {
        if (formFactor() == Plasma::Horizontal || formFactor() == Plasma::Vertical)
        {
            setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred));

            setMaximumHeight(QWIDGETSIZE_MAX);
        }
        else
        {
            setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));

            qreal top, bottom;

            getContentsMargins(0, &top, 0, &bottom);

            setMaximumHeight(m_comboBox->size().height() + top + bottom);
        }

        adjustSize();
    }
}

void RunCommand::activated()
{
    raise();

    m_comboBox->widget()->setFocus();
}

void RunCommand::clearHistory()
{
    KSharedConfig::Ptr sharedConfiguration = KSharedConfig::openConfig("krunnerrc");
    KConfigGroup *configuration = new KConfigGroup(sharedConfiguration, "General");

    configuration->deleteEntry("PastQueries");
    configuration->sync();

    qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->completionObject()->clear();
    qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->clearHistory();
}

void RunCommand::getCommand()
{
    if (qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->lineEdit()->text().isEmpty())
    {
        return;
    }

    runCommand(qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->lineEdit()->text());
}

void RunCommand::runCommand(QString command)
{
    if (m_lastRun.secsTo(QDateTime::currentDateTime()) < 1)
    {
        return;
    }

    m_lastRun = QDateTime::currentDateTime();

    KDisplayManager displayManager;
    KSharedConfig::Ptr sharedConfiguration = KSharedConfig::openConfig("krunnerrc");
    KConfigGroup *configuration = new KConfigGroup(sharedConfiguration, "General");
    KConfig kuriconfig("kuriikwsfilterrc", KConfig::NoGlobals);
    KConfigGroup generalgroup(&kuriconfig, "General");
    QString delimiter = generalgroup.readPathEntry("KeywordDelimiter", QString(":"));
    QString path = QDir::cleanPath(KShell::tildeExpand(command));
    QFileInfo *location = new QFileInfo(path);
    QRegExp webAddress("^((ht|f)tps?://|www\\.).+");
    int index;
    bool commandStatus = true;
    bool addToHistory = true;

    if (command.length() > 3 && command.contains(delimiter))
    {
        index = command.indexOf(delimiter);

        if (index != (command.length() - 1))
        {
            KService::List offers = KServiceTypeTrader::self()->query("SearchProvider", QString("'%1' in Keys").arg(command.left(index)));

            if (!offers.isEmpty())
            {
                QString query = offers.at(0)->property("Query").toString();

                command = query.replace("\\{@}", command.right(command.length() - index - 1));
            }
        }
    }

    KUrl url(command);

    if (command.startsWith("=") || command.startsWith("bin=") || command.startsWith("oct=") || command.startsWith("dec=") || command.startsWith("hex="))
    {
        int base = 10;
        QString result;

        addToHistory = false;

        if (!command.startsWith("="))
        {
            if (command.startsWith("bin="))
            {
                base = 2;
            }
            else if (command.startsWith("oct="))
            {
                base = 8;
            }
            else if (command.startsWith("hex="))
            {
                base = 16;
            }

            command.remove(0, 3);
        }

        command.remove(0, 1).replace(" ", "");

        if (command.contains(KGlobal::locale()->decimalSymbol(), Qt::CaseInsensitive))
        {
            command.replace(KGlobal::locale()->decimalSymbol(), ".", Qt::CaseInsensitive);
        }

        if (command.contains("0x"))
        {
            int position = 0;
            QString hex;

            while ((position = command.indexOf("0x")) > -1)
            {
                hex.clear();

                for (int i = (position + 2); i < command.size(); ++i)
                {
                    if (((command[i] <= '9') && (command[i] >= '0')) || ((command[i] <= 'f') && (command[i] >= 'a')) || ((command[i] <= 'F') && (command[i] >= 'A')))
                    {
                        hex.append(command[i]);
                    }
                    else
                    {
                        break;
                    }
                }

                command.replace("0x" + hex, QString::number(hex.toInt(NULL, 16)));
            }
        }

        command.replace(QRegExp("([a-zA-Z]+)"), "Math.\\1");

        QScriptEngine engine;
        QScriptValue value = engine.evaluate(command);

        if (!value.isError())
        {
            result = ((base == 16)?"0x":"") + QString::number(value.toString().toInt(), base).toUpper();

            qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->lineEdit()->setText(result);
        }
        else
        {
            commandStatus = false;
        }
    }
    else if (location->exists())
    {
        command = path;

        new KRun(path, NULL);
    }
    else if ((url.protocol() != "file" && KProtocolInfo::isKnownProtocol(url.protocol())) || webAddress.exactMatch(command))
    {
        if (url.protocol().isEmpty())
        {
            index = command.indexOf('/');

            url.clear();
            url.setHost(command.left(index));

            if (index != -1)
            {
                url.setPath(command.mid(index));
            }

            url.setProtocol("http");

            command.prepend("http://");
        }

        KToolInvocation::invokeBrowser(url.url());
    }
    else if (command.startsWith("mailto:"))
    {
        KToolInvocation::invokeMailer(command.remove(0, 7), NULL);
    }
    else if (command == "logout")
    {
        KWorkSpace::requestShutDown(KWorkSpace::ShutdownConfirmDefault, KWorkSpace::ShutdownTypeNone);
    }
    else if (command == "shutdown")
    {
        KWorkSpace::requestShutDown(KWorkSpace::ShutdownConfirmDefault, KWorkSpace::ShutdownTypeHalt);
    }
    else if (command == "restart" || command == "reboot")
    {
        KWorkSpace::requestShutDown(KWorkSpace::ShutdownConfirmDefault, KWorkSpace::ShutdownTypeReboot);
    }
    else if (command == "switch" && KAuthorized::authorizeKAction("start_new_session") && displayManager.isSwitchable() && displayManager.numReserve() > 0)
    {
        displayManager.startReserve();
    }
    else if (command == "lock")
    {
        KRun::runCommand("dbus-send --print-reply --dest=org.freedesktop.ScreenSaver /ScreenSaver org.freedesktop.ScreenSaver.Lock", NULL);
    }
    else
    {
        if (command.startsWith("$"))
        {
            KToolInvocation::invokeTerminal(command.remove(0, 1));
        }
        else
        {
            QString binaryName = KRun::binaryName(command, true);

            if (!QFile(binaryName).exists() && KStandardDirs::findExe(binaryName).isEmpty())
            {
                commandStatus = false;
            }
            else
            {
                commandStatus = KRun::runCommand(command, NULL);
            }
        }
    }

    if (commandStatus)
    {
        if (addToHistory)
        {
            qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->addToHistory(command);
            qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->clearEditText();

            configuration->writeEntry("PastQueries", qobject_cast<KHistoryComboBox*>(m_comboBox->widget())->historyItems());
            configuration->sync();
        }
    }
    else
    {
        m_comboBox->widget()->setStyleSheet("QLineEdit, QComboBox {background: pink;}");
    }
}

void RunCommand::textChanged(QString text)
{
    Q_UNUSED(text)

    m_comboBox->widget()->setStyleSheet("QLineEdit, QComboBox {background: auto;}");
}

#include "RunCommand.moc"
